# =====================================
# generator=datazen
# version=3.1.0
# hash=667a45b6e3039f998be6e2945ea7cc28
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "A collection of core Python utilities."
PKG_NAME = "vcorelib"
VERSION = "1.2.5"

# vcorelib-specific content.
DEFAULT_INDENT = 2
DEFAULT_ENCODING = "utf-8"
